/*
 * Realtek RTL2832 demod driver
 *
 * Copyright (c) 2008 Realtek
 * Copyright (c) 2008 Jan Hoogenraad, Barnaby Shearer, Andy Hasper
 * Copyright (C) 2009 Antti Palosaari <crope@iki.fi>
 * Copyright (c) 2011 Maxim Levitsky <maximlevitsky@gmail.com>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include "dvb_frontend.h"
#include "rtl2832.h"


struct rtl2832_dev {
	struct i2c_adapter *i2c;
	struct dvb_frontend frontend;
	struct rtl2832_config config;

	u8 current_page;
	bool configured;

	/* TODO: more private info here */
};

/* write few bytes to demod via i2c */
static int rtl2832_write_i2c(struct rtl2832_dev *dev, u8 reg, u8 *buf, int len)
{
	u8 tmpbuffer[len + 1];

	struct i2c_msg msg = {
		.addr = dev->config.i2c_address,
		.flags = 0, /*write */
		.len = ARRAY_SIZE(tmpbuffer),
		.buf = tmpbuffer,
	};

	tmpbuffer[0] = reg;
	memcpy(tmpbuffer + 1, buf, len);

	if (i2c_transfer(dev->i2c, &msg, 1) != 1)
		return -EIO;
	return 0;
}

/* read few bytes from demod via i2c */
static int rtl2832_read_i2c(struct rtl2832_dev *dev, u8 reg, u8 *buf, int len)
{
	struct i2c_msg msg[] = {
		{
			.addr = dev->config.i2c_address,
			.flags = 0, /*write */
			.len = 1,
			.buf = &reg,
		},
		{
			.addr = dev->config.i2c_address,
			.flags = I2C_M_RD,
			.len = len,
			.buf = buf,
		},
	};

	if (i2c_transfer(dev->i2c, msg, 2) != 2)
		return -EIO;
	return 0;
}

/* read a demod register. register is in packed format
 			and thus one of registers declared in header */
static int rtl2832_read_reg(struct rtl2832_dev *dev, u32 reg, u32 *retval)
{
	u8 page   = (reg >>  0) & 0xFF;
	u8 offset = (reg >>  8) & 0xFF;
	u8 msb    = (reg >> 16) & 0xFF;
	u8 lsb    = (reg >> 24) & 0xFF;

	u32 buffer = 0xCCCCCCCC;
	u8 len = DIV_ROUND_UP(msb + 1, 8);

	deb_reg("%s: page: %02x offset: %02x msb: %02x lsb %02x\n",
		__func__, page, offset, msb, lsb);


	WARN_ON(msb > 31 || lsb > msb || lsb > 7);

	if (page != dev->current_page) {
		if (rtl2832_write_i2c(dev, 0, &page, 1))
			goto error;
		dev->current_page = page;
	}

	if(rtl2832_read_i2c(dev, offset, (u8 *)&buffer, len))
		goto error;

	buffer = le32_to_cpu(buffer);
	deb_reg("read value: %08x\n", buffer);

	buffer &= MASK(msb);
	buffer >>= lsb;

	deb_reg("return value: %08x\n", buffer);


	*retval = buffer;
	return 0;
error:
	return -EIO;
}

/* write a demod register. register is in packed format
 			and thus one of registers declared in header */
static int rtl2832_write_reg(struct rtl2832_dev *dev, u32 reg, u32 value)
{
	u8 page   = (reg >>  0) & 0xFF;
	u8 offset = (reg >>  8) & 0xFF;
	u8 msb    = (reg >> 16) & 0xFF;
	u8 lsb    = (reg >> 24) & 0xFF;

	u32 buffer = 0xCCCCCCCC;
	u8 len = DIV_ROUND_UP(msb + 1, 8);

	deb_reg("%s: page: %02x offset: %02x msb: %02x lsb %02x value %08x\n",
		__func__, page, offset, msb, lsb, value);


	WARN_ON(msb > 31 || lsb > msb || lsb > 7);

	if (page != dev->current_page) {
		if (rtl2832_write_i2c(dev, 0, &page, 1))
			goto error;
		dev->current_page = page;
	}


	if (lsb || (msb + 1) & 0x07) {
		if(rtl2832_read_i2c(dev, offset, (u8 *)&buffer, len))
			goto error;

		buffer = le32_to_cpu(buffer);

		deb_reg("read value: %08x\n", buffer);

		/* clear the bits we are going to write */
		deb_reg("clear mask: %08x\n",
			~(MASK(msb) & ~MASK(lsb - 1)));

		buffer &= ~(MASK(msb) & ~MASK(lsb - 1));

		deb_reg("read value after bit clean: %08x\n", buffer);

	} else
		buffer = 0;

	deb_reg("setting mask: %08x\n", ((value << lsb) & MASK(msb)));
	buffer |= ((value << lsb) & MASK(msb));


	deb_reg("writing: %08x\n", buffer);
	buffer = cpu_to_le32(buffer);

	if (rtl2832_write_i2c(dev, offset, (u8*) &buffer, len))
		goto error;

	return 0;
error:
	return -EIO;
}

/* write bunch of registers at same time */
static int rtl2832_upload_register_table(struct rtl2832_dev *dev,
				struct rtl2832_register_table_entry *table)
{
	while (table->reg) {
		if (rtl2832_write_reg(dev, table->reg, table->value))
			goto error;
		table++;
	}

	return 0;
error:
	return -EIO;
}

extern struct rtl2832_register_table_entry*
			tuner_register_tables[RTL2832_TUNER_LISTSIZE];
extern struct rtl2832_register_table_entry rtl2832_common_regtable[];
extern struct rtl2832_register_table_entry rtl2832_ts_parallel_regtable[];
extern struct rtl2832_register_table_entry rtl2832_ts_serial_regtable[];


/* Initialization (on frontend open usually) */
int rtl2832_init(struct dvb_frontend* fe)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* don't allow opens untill we got a tuner */
	if (!dev->configured)
		goto error;

	/* TODO: that is not complete yet */


	if (rtl2832_upload_register_table(dev, rtl2832_common_regtable))
		goto error;

	if (rtl2832_upload_register_table(dev, dev->config.ts_parallel ?
				rtl2832_ts_parallel_regtable :
				rtl2832_ts_serial_regtable))
		goto error;

	if (tuner_register_tables[dev->config.tuner_type])
		if (rtl2832_upload_register_table(dev,
				tuner_register_tables[dev->config.tuner_type]))
			goto error;

	return 0;
error:
	deb_info("%s: error\n", __func__);
	return -ENODEV;
}

/* Housekeeping on frontend device close */
void rtl2832_release(struct dvb_frontend* fe)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return;
}


/* Set demod to low power state */
int rtl2832_sleep(struct dvb_frontend* fe)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return 0;
}

/* Main tune function */
static int rtl2832_set_parameters(struct dvb_frontend *fe,
			       struct dvb_frontend_parameters *fep)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return 0;
}


static int rtl2832_get_parameters(struct dvb_frontend *fe,
			       	struct dvb_frontend_parameters *fep)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return 0;
}

static int rtl2832_get_tune_settings(struct dvb_frontend *fe,
				    struct dvb_frontend_tune_settings *tune)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return 0;
}


static int rtl2832_read_status(struct dvb_frontend* fe, fe_status_t* status)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return 0;
}

static int rtl2832_read_ber(struct dvb_frontend* fe,u32* ber)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return 0;
}


static int rtl2832_read_signal_strength(struct dvb_frontend* fe, u16* strength)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return 0;
}

static int rtl2832_read_snr(struct dvb_frontend* fe, u16* snr)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return 0;
}

static int rtl2832_read_signal_quality(struct dvb_frontend* fe, u32* quality)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;

	/* TODO */
	return 0;
}

/* Open/close tuner i2c gate */
static int rtl2832_i2c_gate_ctrl(struct dvb_frontend *fe, int enable)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;
	return rtl2832_write_reg(dev, DVBT_IIC_REPEAT, !!enable);
}

static struct dvb_frontend_ops rtl2832_ops = {
	.info = {
		.name			   = "Realtek DVB-T RTL2832",
		.type			   = FE_OFDM,
		.frequency_min	  = 50000000,
		.frequency_max	  = 862000000,
		.frequency_stepsize = 166667,

		.caps = FE_CAN_FEC_1_2 | FE_CAN_FEC_2_3 | FE_CAN_FEC_3_4 |
			FE_CAN_FEC_5_6 | FE_CAN_FEC_7_8 | FE_CAN_FEC_AUTO |

			FE_CAN_QPSK | FE_CAN_QAM_16 |
			FE_CAN_QAM_64 | FE_CAN_QAM_AUTO |

			FE_CAN_TRANSMISSION_MODE_AUTO |
			FE_CAN_GUARD_INTERVAL_AUTO |
			FE_CAN_HIERARCHY_AUTO | FE_CAN_RECOVER |
			FE_CAN_MUTE_TS
	},

	.init =				rtl2832_init,
	.release =			rtl2832_release,

	.sleep =			rtl2832_sleep,

	.set_frontend =			rtl2832_set_parameters,
	.get_frontend =			rtl2832_get_parameters,
	.get_tune_settings =		rtl2832_get_tune_settings,

	.read_status =			rtl2832_read_status,
	.read_ber =				rtl2832_read_ber,
	.read_signal_strength =		rtl2832_read_signal_strength,
	.read_snr =			rtl2832_read_snr,
	.read_ucblocks =		rtl2832_read_signal_quality,

	.i2c_gate_ctrl = 		rtl2832_i2c_gate_ctrl,
};

/* called by bridge driver to obtain an instance of the demod */
struct dvb_frontend *rtl2832_attach(struct i2c_adapter *i2c)
{
	struct rtl2832_dev *dev =
		kzalloc(sizeof(struct rtl2832_dev), GFP_KERNEL);

	if (!dev)
		return NULL;

	dev->i2c = i2c;
	dev->current_page = 0xFF;

	memcpy(&dev->frontend.ops, &rtl2832_ops,
					sizeof(struct dvb_frontend_ops));

	dev->frontend.demodulator_priv = dev;
	return &dev->frontend;
}
EXPORT_SYMBOL(rtl2832_attach);

/* called by bridge driver to tell the demod the tuner type */
int rtl2832_set_config(struct dvb_frontend *fe,
					struct rtl2832_config *config)
{
	struct rtl2832_dev *dev = fe->demodulator_priv;
	dev->config = *config;

	/* TODO: do tuner based initialization here */

	dev->configured = true;
	return 0;
}
EXPORT_SYMBOL(rtl2832_set_config);