/*
 * Realtek RTL28XX DVB USB driver
 *
 *    Copyright (c) 2008 Realtek
 *    Copyright (c) 2008 Jan Hoogenraad, Barnaby Shearer, Andy Hasper
 *    Copyright (C) 2009 Antti Palosaari <crope@iki.fi>
 *    Copyright (c) 2011 Maxim Levitsky <maximlevitsky@gmail.com>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Thanks to Realtek for a lot of support we received !
 */

#include "rtl2832u.h"
#include "rtl2832.h"

int r2832u_debug = 1;
module_param_named(debug, r2832u_debug, int, 0644);
MODULE_PARM_DESC(debug, "set debugging level" DVB_USB_DEBUG_STATUS);
DVB_DEFINE_MOD_OPT_ADAPTER_NR(adapter_nr);


/* control endpoint IO wrapper */
/* 'dummy' - for debug will go in future */
static int r2832u_ctrl_msg(struct dvb_usb_device *d,
			u16 index, u16 value, u8* buffer, int len, bool dummy)
{
	int ret;
	unsigned int pipe;
	u8 requesttype, tmpbuf[len];

	if (index & CTL_WRITE) {
		/* write */
		memcpy(tmpbuf, buffer, len);
		requesttype = (USB_TYPE_VENDOR | USB_DIR_OUT);
		pipe = usb_sndctrlpipe(d->udev, 0);
	} else {
		/* read */
		requesttype = (USB_TYPE_VENDOR | USB_DIR_IN);
		pipe = usb_rcvctrlpipe(d->udev, 0);
	}

	if (dummy)
		ret = 0;
	else
		ret = usb_control_msg(d->udev, pipe, 0, requesttype, value,
			index, tmpbuf, len, RTL2831U_USB_TIMEOUT);

	/* read request, copy returned data to return buf */
	if (requesttype == (USB_TYPE_VENDOR | USB_DIR_IN))
		memcpy(buffer, tmpbuf, len);

	if ((ret < 0 && INFO_DEBUG) || (r2832u_debug & XFER_DEBUG)) {
		printk("control message: index = %04x value"
		" = %04x, result = %d\n", index, value, ret);

		print_hex_dump(KERN_INFO, "message contents: ",
			DUMP_PREFIX_NONE, 32, 1, buffer, len, false);

		printk("\n");

		if (ret < 0)
			return ret;
	}


	return 0;
}

/* dword register read */
static int r2832u_read32(struct dvb_usb_device *dev, u16 offset, u32 *retval)
{
	u32 reg;
	int error = r2832u_ctrl_msg(dev, CTL_REG | CTL_REG_SKIP_BYTESWAP,
						offset, (u8*)&reg, sizeof(reg), false);

	*retval = be32_to_cpu(reg);
	return error;
}

/* dword register write */
static int r2832u_write32(struct dvb_usb_device *dev, u16 offset, u32 value)
{
	value = cpu_to_be32(value);

	/* TODO: check CTL_REG_SKIP_BYTESWAP on older firmware */
	return r2832u_ctrl_msg(dev, CTL_REG | CTL_WRITE | CTL_REG_SKIP_BYTESWAP,
				offset, (u8*)&value, sizeof(value), false);
}

/* byte register read */
static int r2832u_read8(struct dvb_usb_device *dev, u16 offset, u8 *value)
{
	/* TODO: check CTL_REG_SKIP_BYTESWAP on older firmware */
	return r2832u_ctrl_msg(dev, CTL_REG | CTL_REG_SKIP_BYTESWAP,
						offset, value, sizeof(*value), false);
}

/* byte register write */
static int r2832u_write8(struct dvb_usb_device *dev, u16 offset, u8 value)
{
	/* TODO: check CTL_REG_SKIP_BYTESWAP on older firmware */
	return r2832u_ctrl_msg(dev,
		CTL_REG | CTL_WRITE | CTL_REG_SKIP_BYTESWAP,
					offset, &value, sizeof(value), false);
}

/* set/clear a bitmask in 8-bit register */
static int r2832u_set_clear_mask8(struct dvb_usb_device *dev,
					u16 offset, u8 mask, bool set)
{
	u8 reg;
	int retval;

	if ((retval = r2832u_read8(dev, offset, &reg)))
		goto error;
	if (set)
		reg |= mask;
	else
		reg &= ~mask;

	retval = r2832u_write8(dev, offset, reg);
error:
	return retval;
}

/* set/clear a bitmask in 32-bit register */
static int r2832u_set_clear_mask32(struct dvb_usb_device *dev,
					u16 offset, u32 mask, bool set)
{
	u32 reg;
	int retval;

	if ((retval = r2832u_read32(dev, offset, &reg)))
		goto error;
	if (set)
		reg |= mask;
	else
		reg &= ~mask;

	retval = r2832u_write32(dev, offset, reg);
error:
	return retval;
}

/* I2C transfer function*/
static int r2832u_i2c_xfer(struct i2c_adapter *adap, struct i2c_msg msg[],
			   					int num)
{
	struct dvb_usb_device *d = i2c_get_adapdata(adap);
	int i = 0;
	int ret = 0;

	if (mutex_lock_interruptible(&d->i2c_mutex) < 0)
		return -EAGAIN;

	while (i < num && !ret) {

		/* i2c register write - single write of more that one byte
			can always be presented as a regiser offset +
			data write */
		if (!(msg[i].flags & I2C_M_RD) && msg[i].len > 1) {
			ret = r2832u_ctrl_msg(d, CTL_I2C | CTL_WRITE,
				(msg[i].buf[0] << 8) + (msg[i].addr << 1),
				 msg[i].buf + 1, msg[i].len-1, false);
			i++;
			continue;
		}

		/* i2c register read - a write of one byte followed
			by read to _same_ i2c address*/
		if (num > i + 1
			&& !(msg[i].flags & I2C_M_RD)
			&& msg[i].len == 1
			&& msg[i+1].flags & I2C_M_RD
			&& msg[i].addr == msg[i+1].addr) {

			ret = r2832u_ctrl_msg(d, CTL_I2C,
				(msg[i].buf[0] << 8) | (msg[i].addr << 1),
				msg[i+1].buf,
				msg[i+1].len, false);

			i += 2;
			continue;
		}

		/* just writing the device address, aka smbbus quick
			command isn't supported by firmware */
		if (!msg[i].len) {
			ret = -EIO;
			continue;
		}

		/* for all other cases use raw i2c access */
		/* TODO: check against older firmware */
		ret = r2832u_ctrl_msg(d, CTL_I2C_RAW |
			((msg[i].flags & I2C_M_RD) ? 0 : CTL_WRITE),
			msg[i].addr << 1 , msg[i].buf, msg[i].len, false);
		i++;
	}

	mutex_unlock(&d->i2c_mutex);
	return ret ? ret : i;
}

static u32 r2832u_i2c_func(struct i2c_adapter *adapter)
{
	return (I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL) & ~I2C_FUNC_SMBUS_QUICK;
}

static struct i2c_algorithm r2832u_i2c_algo = {
	.master_xfer   = r2832u_i2c_xfer,
	.functionality = r2832u_i2c_func,
};

static int r2832u_set_gpio_direction(struct dvb_usb_device *d,
					unsigned int gpio, int direction)
{
	if (direction == GPIO_IN && r2832u_set_clear_mask8(d,
				SYS_GPIO_OUTPUT_ENABLE, 1 << gpio, false))
		goto error;

	if (r2832u_set_clear_mask8(d, SYS_GPIO_DIRECTION,
				1 << gpio, direction == GPIO_IN))
		goto error;

	if (direction == GPIO_OUT && r2832u_set_clear_mask8(d,
				SYS_GPIO_OUTPUT_ENABLE, 1 << gpio, true))
		goto error;

	return 0;
error:
	return -EIO;
}

static int r2832u_set_gpio_output(struct dvb_usb_device *d, int gpio, bool value)
{
	return r2832u_set_clear_mask8(d, SYS_GPIO_OUTPUT_VALUE, 1 << gpio, value);
}

static int r2832u_set_gpio_pad(struct dvb_usb_device *d, int gpio, int pad)
{
	int pad_reg = gpio > 3 ? SYS_GPIO_PAD1 : SYS_GPIO_PAD0;
	u8 reg;

	gpio &= ~0x04;

	if (r2832u_read8(d, pad_reg, &reg))
		goto error;

	reg &= ~(0x03 << (gpio & 0x03));
	reg |= pad << (gpio & 0x03);

	if (r2832u_write8(d, pad_reg, reg))
		goto error;
	return 0;
error:
	return -EIO;
}


/* set settings sutable for sleep or running state */
static int r2832u_power(struct dvb_usb_device *d, bool on)
{
	u32 reg;
	struct r2832u_priv *priv = d->priv;

	/* 1.FULL packet mode(for bulk)
	 * 2.DMA enable.
	 */
	if(r2832u_set_clear_mask32(d, USB_SYSCTL, 0x09, true))
		goto error;

	/* check epa config,
	 * [9-8]:00, 1 transaction per microframe
	 * [7]:1, epa enable
	 * [6-5]:10, bulk mode
	 * [4]:1, device to host
	 * [3:0]:0001, endpoint number
	 */
	if(r2832u_read32(d, USB_EPA_CFG, &reg) || ((reg & 0x03FF) != 0xD1))
		goto error;

	/* read usb speed mode - BTW we can just ask linux for that */
	if (r2832u_read32(d, USB_STAT, &reg))
		goto error;

	priv->high_speed = !!(reg & 0x01);

	/* Set the bulk endpoint maximum packet size */
	if (r2832u_write32(d, USB_EPA_MAXPKT, priv->high_speed ? 0x200 : 0x040))
		goto error;

	/* Set the bulk endpoint fifo size */
	if (r2832u_write32(d, USB_EPA_FIFO_CFG, 0x14))
		goto error;

	/* reset epa fifo:
	 * [9] Reset EPA FIFO
	 * [5] FIFO Flush,Write 1 to flush the oldest TS packet (a 188 bytes block)
	 */
	if (r2832u_write32(d, USB_EPA_CTL, 0x0210))
		goto error;
	if (r2832u_write32(d, USB_EPA_CTL, 0))
		goto error;

	/* Set so called GPIO3. This is somehow related with wakeup */

	/* GPIO3_PAD Pull-HIGH / Output high on GPIO3*/
	if (r2832u_set_gpio_pad(d, 3, 0x02))
		goto error;
	if (r2832u_set_gpio_direction(d, 3, GPIO_OUT))
		goto error;
	if (r2832u_set_gpio_output(d, 3, true))
		goto error;

	/* Set the BTN_WAKEUP_DIS bit in USB_CTRL */
	if (r2832u_set_clear_mask32(d, USB_CTRL, 0x20, true))
		goto error;

	/* Set/clear the 'Suspend Latch' */
	if (r2832u_set_clear_mask8(d, SYS_DEMOD_CTL1, 0x10, !on))
		goto error;

	/* demod PLL setting */
	if (r2832u_set_clear_mask8(d, SYS_DEMOD_CTL, 0x80, on))
		goto error;

	if (on) {

		/* Perform a hardware reset of the demod */
		if (r2832u_set_clear_mask8(d, SYS_DEMOD_CTL, 0x20, false))
			goto error;
		if (r2832u_set_clear_mask8(d, SYS_DEMOD_CTL, 0x20, true))
			goto error;

		msleep(5);
	}

	/* demod: Set ADC_Q and ADC_I setting */
	if (r2832u_set_clear_mask8(d, SYS_DEMOD_CTL, 0x44, on))
		goto error;

	return 0;
error:
	pr_info("%s: error", __func__);
	return -EIO;
}


/* init rtl2832 demod - rtl2836/rtl2840 will be eventually done in other
	function */
static int r2832u_frontend_attach(struct dvb_usb_adapter *adap)
{
	struct dvb_usb_device *d = adap->dev;
	struct dvb_frontend *fe;

	struct rtl2832_config config = {
		.i2c_address = 0x10,
		.crystal_freq = 28800000,
		.ts_parallel = true,
		.periodic_update_hack = false,
		.tuner_type = RTL2832_TUNER_FC0012,
	};

	deb_info("%s:\n", __func__);

	r2832u_power(d, true);
	fe = rtl2832_attach(&adap->dev->i2c_adap);

	/* TODO: will be done at tuner_attach */
	rtl2832_set_config(fe, &config);
	fe->ops.init(fe);

	return 0;
}

/* called by dvb-usb - select tuner */
static int r2832u_tuner_attach(struct dvb_usb_adapter *adap)
{
	deb_info("%s:\n", __func__);
	return 0;
}

/* start/stop DVB stream */
static int r2832u_streaming_ctrl(struct dvb_usb_adapter *adap , int onoff)
{
	deb_info("%s: onoff:%d\n", __func__, onoff);

	/* endpoint A
	 * if turning off, then:
	 * stall epa, set bit 4 to 1
	 * reset epa, set bit 9 to 1
	 */

	if(r2832u_write32(adap->dev, USB_EPA_CTL, onoff ? 0 : 0x0210))
		return -EFAULT;

	return 0;
}

static struct usb_device_id r2832u_table[] = {
	{USB_DEVICE(USB_VID_REALTEK,    USB_PID_RTD2831U)},
	{USB_DEVICE(USB_VID_REALTEK,    USB_PID_RTD2838U)},
	{} /* Terminating entry */
};

//MODULE_DEVICE_TABLE(usb, r2832u_table);

static struct dvb_usb_device_properties r2832u_properties[] = {
	{
		.caps = DVB_USB_IS_AN_I2C_ADAPTER,

		.usb_ctrl = DEVICE_SPECIFIC,
		.no_reconnect = 1,

		.size_of_priv = sizeof(struct r2832u_priv),

		.num_adapters = 1,
		.adapter = {
			{
				.streaming_ctrl  = r2832u_streaming_ctrl,
				.frontend_attach = r2832u_frontend_attach,
				.tuner_attach    = r2832u_tuner_attach,
				.stream = {
					.type = USB_BULK,
					.count = 6,
					.endpoint = 0x81,
					.u = {
						.bulk = {
							.buffersize = 4096,
						}
					}
				},
			}
		},

		.i2c_algo = &r2832u_i2c_algo,

		.num_device_descs = 2,
		.devices = {
			{
				.name = "RTL2831U DVB-T USB2.0 DEVICE",
				.warm_ids = {&r2832u_table[0], NULL},
			},
			{
				.name = "RTL2838U DVB-T USB2.0 DEVICE",
				.warm_ids = {&r2832u_table[1], NULL},
			},
		}
	},
};

static int r2832u_probe(struct usb_interface *intf,
			const struct usb_device_id *id)
{
	int ret = 0;
	struct dvb_usb_device *d = NULL;
	u8 i;
	u8 properties_count = ARRAY_SIZE(r2832u_properties);

	if (intf->cur_altsetting->desc.bInterfaceNumber == 0)
		return -ENODEV;

	deb_info("%s: interface:%d\n", __func__,
		intf->cur_altsetting->desc.bInterfaceNumber);

	for (i = 0; i < properties_count; i++) {
		ret = dvb_usb_device_init(intf, &r2832u_properties[i],
			THIS_MODULE, &d, adapter_nr);
		if (!ret)
			return 0;
		if (ret != -ENODEV) {
			deb_info("%s: failed:%d\n", __func__, ret);
			return ret;
		}
	}

	return -ENODEV;
}

static struct usb_driver r2832u_driver = {
	.name       = "dvb_usb_r2832u",
	.probe      = r2832u_probe,
	.disconnect = dvb_usb_device_exit,
	.id_table   = r2832u_table,
};

/* module stuff */
static int __init r2832u_module_init(void)
{
	int ret;
	deb_info("%s:\n", __func__);
	ret = usb_register(&r2832u_driver);
	if (ret)
		err("usb_register failed:%d", ret);

	return ret;
}

static void __exit r2832u_module_exit(void)
{
	deb_info("%s:\n", __func__);
	usb_deregister(&r2832u_driver);
}

module_init(r2832u_module_init);
module_exit(r2832u_module_exit);

MODULE_DESCRIPTION("Realtek RTL28XXU DVB-T USB2.0 driver");
MODULE_AUTHOR("ChiaLing"); /* Realtek */
MODULE_AUTHOR("Antti Palosaari <crope@iki.fi>");
MODULE_LICENSE("GPL");
